!
! auto-generated by op2_fortran.py
!

program AIRFOIL
  use OP2_FORTRAN_DECLARATIONS
!  use OP2_FORTRAN_HDF5_DECLARATIONS
  use SAVE_SOLN_MODULE
  use ADT_CALC_MODULE
  use RES_CALC_MODULE
  use BRES_CALC_MODULE
  use UPDATE_MODULE
   
  use OP2_CONSTANTS
  use AIRFOIL_SEQ
  use IO
  use, intrinsic :: ISO_C_BINDING

  implicit none

  intrinsic :: sqrt, real

  integer(4) :: iter, k, i

  integer(4), parameter :: maxnode = 9900
  integer(4), parameter :: maxcell = (9702+1)
  integer(4), parameter :: maxedge = 19502

  integer(4), parameter :: iterationNumber = 1000

  integer(4) :: nnode, ncell, nbedge, nedge, niter, qdim
  real(8) :: ncellr

  ! profiling
  real(kind=c_double) :: startTime = 0
  real(kind=c_double) :: endTime = 0

  ! integer references (valid inside the OP2 library) for op_set
  type(op_set) :: nodes, edges, bedges, cells

  ! integer references (valid inside the OP2 library) for pointers between data sets
  type(op_map) :: pedge, pecell, pcell, pbedge, pbecell

  ! integer reference (valid inside the OP2 library) for op_data
  type(op_dat) :: p_bound, p_x, p_q, p_qold, p_adt, p_res

  ! arrays used in data
  integer(4), dimension(:), allocatable, target :: ecell, bound, edge, bedge, becell, cell
  real(8), dimension(:), allocatable, target :: x, q, qold, adt, res, q_part
  real(8), dimension(1:2) :: rms

  integer(4) :: debugiter, retDebug
  real(8) :: datad

  ! for validation
  REAL(KIND=8) :: diff
  integer(4):: ncelli

  ! read set sizes from input file (input is subdivided in two routines as we cannot allocate arrays in subroutines in
  ! fortran 90)
  print *, "Getting set sizes"
  call getSetSizes ( nnode, ncell, nedge, nbedge )

  print *, ncell
  ! allocate sets (cannot allocate in subroutine in F90)
  allocate ( cell ( 4 * ncell ) )
  allocate ( edge ( 2 * nedge ) )
  allocate ( ecell ( 2 * nedge ) )
  allocate ( bedge ( 2 * nbedge ) )
  allocate ( becell ( nbedge ) )
  allocate ( bound ( nbedge ) )

  allocate ( x ( 2 * nnode ) )
  allocate ( q ( 4 * ncell ) )
  allocate ( qold ( 4 * ncell ) )
  allocate ( res ( 4 * ncell ) )
  allocate ( adt ( ncell ) )

  !allocated simply to test op_fetch_data_idx()
  allocate ( q_part ( 4 * ncell ) )

  print *, "Getting data"
  call getSetInfo ( nnode, ncell, nedge, nbedge, cell, edge, ecell, bedge, becell, bound, x, q, qold, res, adt )

  ! OP initialisation
  call op_init_base (0,0)

  print *, "Initialising constants"
  call initialise_flow_field ( ncell, q, res )

  ! declare sets, pointers, datasets and global constants (for now, no new partition info)
  print *, "Declaring OP2 sets"
  call op_decl_set ( nnode, nodes, 'nodes' )
  call op_decl_set ( nedge, edges, 'edges' )
  call op_decl_set ( nbedge, bedges, 'bedges' )
  call op_decl_set ( ncell, cells, 'cells' )

  print *, "Declaring OP2 maps"
  call op_decl_map ( edges, nodes, 2, edge, pedge, 'pedge' )
  call op_decl_map ( edges, cells, 2, ecell, pecell, 'pecell' )
  call op_decl_map ( bedges, nodes, 2, bedge, pbedge, 'pbedge' )
  call op_decl_map ( bedges, cells, 1, becell, pbecell, 'pbecell' )
  call op_decl_map ( cells, nodes, 4, cell, pcell, 'pcell' )

  print *, "Declaring OP2 data"
  call op_decl_dat ( bedges, 1, 'integer' ,bound, p_bound, 'p_bound')
  call op_decl_dat ( nodes, 2, 'real(8)',x, p_x, 'p_x' )
  call op_decl_dat ( cells, 4, 'real(8)', q, p_q, 'p_q' )
  call op_decl_dat ( cells, 4, 'real(8)', qold, p_qold, 'p_qold' )
  call op_decl_dat ( cells, 1, 'real(8)', adt, p_adt, 'p_adt' )
  call op_decl_dat ( cells, 4, 'real(8)', res, p_res, 'p_res' )


  print *, "Declaring OP2 constants"
       
  !call op_dump_to_hdf5("new_grid_out.h5")
  !call op_fetch_data_hdf5_file(p_x, "new_grid_out.h5")

  call op_partition ('PTSCOTCH','KWAY', edges, pecell, p_x)

  ncelli  = op_get_size(cells)
  ncellr = real(ncelli)

  ! start timer
  call op_timers ( startTime )

  ! main time-marching loop

  do niter = 1, iterationNumber

     call save_soln_host(&
                      & "save_soln",cells,  &
                      & op_arg_dat(p_q,-1,OP_ID,4,"real(8)",OP_READ),  &
                      & op_arg_dat(p_qold,-1,OP_ID,4,"real(8)",OP_WRITE))


    ! predictor/corrector update loop

    do k = 1, 2

      ! calculate area/timstep
      call adt_calc_host(&
                       & "adt_calc",cells,  &
                       & op_arg_dat(p_x,1,pcell,2,"real(8)",OP_READ),  &
                       & op_arg_dat(p_x,2,pcell,2,"real(8)",OP_READ),  &
                       & op_arg_dat(p_x,3,pcell,2,"real(8)",OP_READ),  &
                       & op_arg_dat(p_x,4,pcell,2,"real(8)",OP_READ),  &
                       & op_arg_dat(p_q,-1,OP_ID,4,"real(8)",OP_READ),  &
                       & op_arg_dat(p_adt,-1,OP_ID,1,"real(8)",OP_WRITE))


      ! calculate flux residual
      call res_calc_host(&
                       & "res_calc",edges,  &
                       & op_arg_dat(p_x,1,pedge,2,"real(8)",OP_READ),  &
                       & op_arg_dat(p_x,2,pedge,2,"real(8)",OP_READ),  &
                       & op_arg_dat(p_q,1,pecell,4,"real(8)",OP_READ),  &
                       & op_arg_dat(p_q,2,pecell,4,"real(8)",OP_READ),  &
                       & op_arg_dat(p_adt,1,pecell,1,"real(8)",OP_READ),  &
                       & op_arg_dat(p_adt,2,pecell,1,"real(8)",OP_READ),  &
                       & op_arg_dat(p_res,1,pecell,4,"real(8)",OP_INC),  &
                       & op_arg_dat(p_res,2,pecell,4,"real(8)",OP_INC))


      call bres_calc_host(&
                       & "bres_calc",bedges,  &
                       & op_arg_dat(p_x,1,pbedge,2,"real(8)",OP_READ),  &
                       & op_arg_dat(p_x,2,pbedge,2,"real(8)",OP_READ),  &
                       & op_arg_dat(p_q,1,pbecell,4,"real(8)",OP_READ),  &
                       & op_arg_dat(p_adt,1,pbecell,1,"real(8)",OP_READ),  &
                       & op_arg_dat(p_res,1,pbecell,4,"real(8)",OP_INC),  &
                       & op_arg_dat(p_bound,-1,OP_ID,1,"integer(4)",OP_READ))


      ! update flow field

      rms(1:2) = 0.0

      call update_host(&
                       & "update",cells,  &
                       & op_arg_dat(p_qold,-1,OP_ID,4,"real(8)",OP_READ),  &
                       & op_arg_dat(p_q,-1,OP_ID,4,"real(8)",OP_WRITE),  &
                       & op_arg_dat(p_res,-1,OP_ID,4,"real(8)",OP_RW),  &
                       & op_arg_dat(p_adt,-1,OP_ID,1,"real(8)",OP_READ),  &
                       & op_arg_gbl(rms,2,"real(8)",OP_INC))


    end do ! internal loop

    rms(2) = sqrt ( rms(2) / ncellr )

    if (op_is_root() .eq. 1) then
      if (mod(niter,100) .eq. 0) then
        write (*,*) niter,"  ",rms(2)
      end if
      if ((mod(niter,1000) .eq. 0) .AND. (ncelli == 720000) ) then
        diff=ABS((100.0_8*(rms(2)/0.0001060114637578_8))-100.0_8)
        !write (*,*) niter,"  ",rms(2)
        WRITE(*,'(a,i,a,e16.7,a)')"Test problem with ", ncelli , &
        & " cells is within ",diff,"% of the expected solution"
        if(diff.LT.0.00001) THEN
          WRITE(*,*)"This test is considered PASSED"
        else
          WRITE(*,*)"This test is considered FAILED"
        endif
      end if
    end if

  end do ! external loop

!  call op_fetch_data(p_q,q)

!  call op_fetch_data_idx(p_q,q_part, 1, ncell)

  call op_timers ( endTime )
  call op_timing_output ()
  write (*,*) 'Max total runtime =',endTime-startTime,'seconds'

end program AIRFOIL
